@@ -1,4 +1,4 @@
-import sys
+import sys, json
 from typing import Dict
 import numpy as np
 import collections
@@ -20,6 +20,13 @@ try:
 except ModuleNotFoundError:
     Redis = None
 
+try:
+    from elasticsearch import Elasticsearch
+    from elasticsearch.helpers import bulk
+    available_engines.add("elastic")
+except ModuleNotFoundError:
+    Elasticsearch = None
+
 
 class BaseIndex:
     def __init__(self, metric:str, dim:int):
@@ -45,19 +52,38 @@ class BaseIndex:
                 index.init()
         self.fitted = True
     def add_items(self, data, ids=None, partition=None):
+        # Support for adding items to multiple partitions
+        if type(partition)==np.str_:
+            partition = str(partition)
+        if hasattr(partition, "__iter__") and type(partition)!=str:
+            data = np.array(data)
+            ids = np.array(ids)
+            ps = set(partition)
+            partition = np.array(partition)
+            for p in ps:
+                self.add_items(data[partition==p], ids[partition==p], p)
+            return
+        # Support for adding items to a single partition
         if ids is None:
             ids = list(range(len(self.items[partition]),len(self.items[partition])+len(data)))
+        elif type(ids)==np.ndarray:
+            ids = ids.tolist()
         if hasattr(self.cls, 'add_items'):
             self.ids[partition].extend(ids)
-            nids = [self.ids[partition].index(i) for i in ids]
-            self.indices[partition].add_items(data, nids)
+            self.indices[partition].add_items(data)
         else:
             self.items[partition].extend(data)
             self.indices[partition]=None
+            n_before = len(set(self.ids[partition]))
             self.ids[partition].extend(ids)
+            n_after = len(set(self.ids[partition]))
+            if n_after != n_before+len(ids):
+                raise ValueError("ids must be unique")
         self.fitted = False
 
     def get_items(self, ids=None,partition=None):
+        if type(ids)==str:
+            raise SyntaxError("ids must be a list")
         if hasattr(self.indices[partition], "get_items"):
             return self.indices[partition].get_items(ids)
         return [self.items[partition][self.ids[partition].index(i)] for i in ids]
@@ -77,12 +103,19 @@ class BaseIndex:
         else:
             scores, names = [], []
             for p in self.indices:
-                scores_p, idx_p = self.indices[p].search(data, k)
+                try:
+                    scores_p, idx_p = self.indices[p].search(data, k)
+                except ValueError:
+                    continue
                 if len(scores_p)==0:
                     continue
                 scores_p, idx_p = scores_p[0], idx_p[0]
                 scores.extend(scores_p)
                 names.extend([self.ids[p][i] for i in idx_p])
+            if len(names)>k:
+                sorter = np.argsort(scores)
+                scores = np.array(scores)[sorter].tolist()[:k]
+                names = np.array(names)[sorter].tolist()[:k]
         return scores, names
     def __repr__(self):
         partitions = ",".join(map(str, self.items.keys()))
@@ -90,70 +123,23 @@ class BaseIndex:
     def __str__(self):
         return self.__repr__()
     def __len__(self):
-        len = 0
+        length = 0
         for p in self.items:
-            len += len(self.items[p])
-        return len
+            length += len(self.items[p])
+        return length
     def  __itemgetter__(self, item):
         for p in self.items:
             if item in self.ids[p]:
                 return self.items[p][self.ids[p].index(item)]
         return None
 
-class FaissIndexUnpartitioned:
-    def __init__(self, metric:str, dim:int, index_factory:str, **kwargs):
-        self.dim = dim
-        self.metric = metric
-        if index_factory == '':
-            index_factory = 'Flat'
-        if metric in ['ip', 'cosine']:
-            self.index = faiss.index_factory(dim, index_factory, faiss.METRIC_INNER_PRODUCT)
-            if metric == 'cosine':
-                sys.stderr.write("cosine is not supported by faiss, falling back to dot\n")
-        elif metric == 'l2':
-            self.index = faiss.index_factory(dim, index_factory, faiss.METRIC_L2)
-        else:
-            raise TypeError(str(metric) + " is not supported")
-        self.index = faiss.IndexIDMap2(self.index)
-
-    def __len__(self):
-        return self.get_current_count()
-
-    def  __itemgetter__(self, item):
-        return self.index.reconstruct(int(item))
-
-    def init(self):
-        pass
-    
-    def add_items(self, data, ids):
-        data = np.array(data).astype(np.float32)
-        self.index.add_with_ids(data, np.array(ids, dtype='int64'))
-
-    def get_items(self, ids):
-        # recmap = {k:i for i,k in enumerate(faiss.vector_to_array(self.index.id_map))}
-        # return np.vstack([self.index.reconstruct(recmap[v]) for v in ids])
-        return np.vstack([self.index.reconstruct(int(v)) for v in ids])
-
-    def get_max_elements(self):
-        return -1
-
-    def get_current_count(self):
-        return self.index.ntotal
-
-    def search(self, data, k=1):
-        return self.index.search(np.array(data).astype(np.float32),k)
-
-    def save_index(self, fname):
-        return faiss.write_index(self.index, fname)
-
-    def load_index(self, fname):
-        self.index = faiss.read_index(fname)
-
-
 class SciKitIndexUnpartitioned(NearestNeighbors):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
     def search(self, data, k=1):
+        if k>self.n_neighbors:
+            sys.stderr.write("Warning: k is larger than n_neighbors, falling back to n_neighbors\n")
+            k = self.n_neighbors
         return super().kneighbors(data, k, return_distance=True)
 
 class SciKitIndex(BaseIndex):
@@ -173,6 +159,7 @@ class RedisIndex(BaseIndex):
         self.max_elements = max_elements
         self.ef_construction = ef_construction
         self.M = M
+        self.partitions = set()
         if kwargs.get("index_name") is None:
             self.index_name = "idx"
         else:
@@ -223,6 +210,8 @@ class RedisIndex(BaseIndex):
 
     def search(self, data, k=1,partition=None):
         """Search the nearest neighbors of the given vectors, and a given partition."""
+        if partition is not None and partition not in self.partitions:
+            raise Exception(f"Partition {partition} does not exist")
         query_vector = np.array(data).astype(np.float32).tobytes()
         #prepare the query
         p = "(@partition:{"+partition+"})" if partition is not None else "*"
@@ -237,13 +226,27 @@ class RedisIndex(BaseIndex):
 
     def add_items(self, data, ids=None, partition=None):
         """Add items and ids to the index, if a partition is not defined it defaults to NONE"""
+        # Support for adding items to multiple partitions
+        if type(partition)==np.str_:
+            partition = str(partition)
+        if hasattr(partition, "__iter__") and type(partition)!=str:
+            data = np.array(data)
+            ids = np.array(ids)
+            ps = set(partition)
+            partition = np.array(partition)
+            for p in ps:
+                self.add_items(data[partition==p], ids[partition==p], p)
+            return
+        # Support for adding items to a single partition
         self.pipe = self.redis.pipeline(transaction=False)
         if partition is None:
             partition="NONE"
+        self.partitions.add(partition)
         for datum, id in zip(data, ids):
             key='item:'+ str(id)
             emb = np.array(datum).astype(np.float32).tobytes()
             self.pipe.hset(key,mapping={"embedding": emb, "item_id": str(id), "partition":partition})
+    
                 
         self.pipe.execute()
         self.pipe = None
@@ -296,10 +299,12 @@ class RedisIndex(BaseIndex):
         ret = self.redis.lrange("user:"+str(user_id), 0, -1)
         return [dict(zip(self.user_keys,x.decode().split('|'))) for x in ret]
     
-    def set_vector(self, key, arr, prefix="vec:"):
+    def set_vector(self, key, arr, prefix="vec:", ttl:int=60*60*24):
         """Sets a numpy array as a vector in redis"""
         emb = np.array(arr).astype(np.float32).tobytes()
         self.redis.set(prefix+str(key), emb)
+        if ttl:
+            self.redis.expire(prefix+str(key), ttl)
         return self
     
     def get_vector(self, key, prefix="vec:"):
@@ -326,10 +331,190 @@ class RedisIndex(BaseIndex):
         """Get Redis info as dict"""
         return self.redis.ft(self.index_name).info()
 
+class ElasticIndex(BaseIndex):
+    def __init__(self, metric:str, dim:int, elastic_credentials,index_name="vecsim",M=16,ef_construction=100,**kwargs):
+        self.es = Elasticsearch(**elastic_credentials)
+        self.index_name = index_name
+        self.partitions = set()
+        if metric=="ip":
+            metric="dot_product"
+        elif metric=="l2":
+            metric="l2_norm"
+        mappings =  {
+            "properties": {
+                "vec": {
+                    "type": "dense_vector",
+                    "dims": dim,
+                    "index": True,
+                    "similarity": metric,
+                    "index_options": {
+                        "type": "hnsw",
+                        "m": M,
+                        "ef_construction": ef_construction
+                    }
+                },
+                "partition" : {
+                    "type": "text",
+                    "analyzer": "keyword"
+                    },
+                "id": {
+                    "type": "text",
+                    "analyzer": "keyword"
+                },
+                },
+                # "text" : {
+                #     "type" : "keyword"
+                # }
+        }
+            
+        self.es.indices.create(index=index_name, ignore=400, body={"mappings":mappings})
+    
+    def add_items(self, data, ids=None, partition=None):
+        # Support for adding items to multiple partitions
+        if type(partition)==np.str_:
+            partition = str(partition)
+        if hasattr(partition, "__iter__") and type(partition)!=str:
+            data = np.array(data)
+            ids = np.array(ids)
+            ps = set(partition)
+            partition = np.array(partition)
+            for p in ps:
+                self.add_items(data[partition==p], ids[partition==p], p)
+            return
+        # Support for adding items to a single partition
+        self.partitions.add(partition)
+        bulk(self.es, [{
+                "_index": self.index_name,
+                "_source": {"vec": [float(x) for x in datum], "id": str(id), "partition":partition},
+            } for datum, id in zip(data, ids)])
+    
+    def search(self, data, k=1,partition=None, num_candidates=50):
+        if partition is not None and partition not in self.partitions:
+            raise Exception(f"Partition {partition} does not exist")
+        if partition:
+            query = {"query": {"bool": {"must": [{"match": {"partition": partition}}, {"knn": {"field":"vec","query_vector": [float(x) for x in data], "k": k, "num_candidates": num_candidates}}]}}}
+        else:
+            query = {"knn": {"field":"vec","query_vector": [float(x) for x in data], "k": k, "num_candidates": num_candidates}}
+        res = self.es.search(index=self.index_name, body=query)
+        ids = [x["_source"]["id"] for x in res["hits"]["hits"]]
+        scores = [x["_score"] for x in res["hits"]["hits"]]
+        return scores, ids
+
+    def init(self, **kwargs):
+        pass
+
+    def get_items(self, ids=None):
+        """Get items by id"""
+        if ids is None:
+            ids = []
+        if not isinstance(ids, list):
+            ids = [ids]
+        res = self.es.mget(index=self.index_name, body={"ids": ids})
+        return [np.array(x["_source"]["vec"]) for x in res["docs"] if x["found"]]
+
+    def get_current_count(self):
+        """Get number of items in index"""
+        return self.es.count(index=self.index_name)["count"]
+    
+    def __len__(self):
+        return self.get_current_count()
+
+    def  __itemgetter__(self, item):
+        return super().get_items([item])[0]
+
+
+
+## Faiss
+
+class FaissBase:
+    def add_items(self, data):
+        data = np.array(data).astype(np.float32)
+        if self.metric=='cosine':
+            faiss.normalize_L2(data)
+        if hasattr(self.index, "is_trained") and not self.index.is_trained:
+            self.index.train(data)
+        # self.data = data
+        self.index.add(data)
+
+    def init(self):
+        pass
+
+    def get_current_count(self):
+        return self.index.ntotal
+
+    def search(self, data, k=1):
+        data = np.array(data).astype(np.float32)
+        if self.metric=='cosine':
+            faiss.normalize_L2(data)
+            # data/=np.linalg.norm(data,axis=1)
+        return self.index.search(data,k)
+
+    def save_index(self, fname):
+        return faiss.write_index(self.index, fname)
+
+    def load_index(self, fname):
+        self.index = faiss.read_index(fname)
+    def __len__(self):
+        return self.get_current_count()
+class FaissFlatIndexUnpartitioned(FaissBase):
+    def __init__(self, metric:str, dim:int):
+        self.dim = dim
+        self.metric = metric
+        if metric in ['ip', 'dot']:
+            self.index = faiss.IndexFlatIP(self.dim)
+        elif metric == 'cosine':
+            self.index = faiss.IndexFlatIP(self.dim)
+        elif metric == 'l2':
+            self.index = faiss.IndexFlatL2(self.dim)
+        else:
+            raise TypeError(str(metric) + " is not supported")
+
+class FaissIVFPQIndexUnpartitioned(FaissFlatIndexUnpartitioned):
+    def __init__(self, metric:str, dim:int, nlist:int=256, m:int=8, nbits:int=8):
+        super().__init__(metric, dim)
+        self.index = faiss.IndexIVFPQ(self.index, self.dim, nlist, m, nbits)
+
+
+class FaissIndexFactoryUnpartitioned(FaissBase):
+    def __init__(self, metric:str, dim:int, index_factory:str, **kwargs):
+        self.dim = dim
+        self.metric = metric
+        if index_factory == '':
+            index_factory = 'Flat'
+        if metric in ['ip', 'dot', 'cosine']:
+            self.index = faiss.index_factory(dim, index_factory, faiss.METRIC_INNER_PRODUCT)
+        elif metric == 'l2':
+            self.index = faiss.index_factory(dim, index_factory, faiss.METRIC_L2)
+        else:
+            raise TypeError(str(metric) + " is not supported")
+
+    def  __itemgetter__(self, item):
+        return self.index.reconstruct(int(item))
+
+    def get_items(self, ids):
+        # recmap = {k:i for i,k in enumerate(faiss.vector_to_array(self.index.id_map))}
+        # return np.vstack([self.index.reconstruct(recmap[v]) for v in ids])
+        return np.vstack([self.index.reconstruct(int(v)) for v in ids])
+
+
 
 class FaissIndex(BaseIndex):
     def __init__(self, metric:str, dim:int, index_factory:str, **kwargs):
         BaseIndex.__init__(self, metric, dim)
-        self.cls = FaissIndexUnpartitioned
+        self.cls = FaissIndexFactoryUnpartitioned
         self.cls_args={"metric":self.metric,"dim":self.dim,"index_factory":index_factory}
         self.indices = collections.defaultdict(lambda:self.cls(**self.cls_args))
+
+class FaissFlatIndex(BaseIndex):
+    def __init__(self, metric:str, dim:int, **kwargs):
+        BaseIndex.__init__(self, metric, dim)
+        self.cls = FaissFlatIndexUnpartitioned
+        self.cls_args={"metric":self.metric,"dim":self.dim}
+        self.indices = collections.defaultdict(lambda:self.cls(**self.cls_args))
+
+class FaissIVFPQIndex(BaseIndex):
+    def __init__(self, metric:str, dim:int, **kwargs):
+        BaseIndex.__init__(self, metric, dim)
+        self.cls = FaissIVFPQIndexUnpartitioned
+        self.cls_args={"metric":self.metric,"dim":self.dim}
+        self.indices = collections.defaultdict(lambda:self.cls(**self.cls_args))
\ No newline at end of file
